unit FindEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, DBTables, DB, DBCtrls, DsgnIntf, MyProc;

type
  TEditCharCase = (ecNormal, ecUpperCase, ecLowerCase);
  TNextRecType = (nrNextRecord, nrNextValue);

  TFindEvent = procedure(Sender: TObject; IsFind: Boolean) of object;

  TCustomFindEdit = class(TWinControl)
  private
    FMaxLength: Integer;
    FBorderStyle: TBorderStyle;
    FPasswordChar: Char;
    FReadOnly: Boolean;
    FAutoSize: Boolean;
    FAutoSelect: Boolean;
    FHideSelection: Boolean;
    FOEMConvert: Boolean;
    FCharCase: TEditCharCase;
    FCreating: Boolean;
    FModified: Boolean;
    FOnChange: TNotifyEvent;
    FDataSet: TDataSet;
    FTableIndex: string;
    FField: string;
    FNewText: string;
    FFindEvent: TFindEvent;
    FCaretPos: Integer;
    FBtnDownX: Integer;
    FPosition: Integer;
    FIsFind: Boolean;
    FNextRecType: TNextRecType;
    procedure AdjustHeight;
    procedure DefaultHandler(var Message); override;
    function GetModified: Boolean;
    procedure SetAutoSize(Value: Boolean);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetCharCase(Value: TEditCharCase);
    procedure SetHideSelection(Value: Boolean);
    procedure SetMaxLength(Value: Integer);
    procedure SetModified(Value: Boolean);
    procedure SetOEMConvert(Value: Boolean);
    procedure SetPasswordChar(Value: Char);
    procedure SetReadOnly(Value: Boolean);
    procedure SetSelText(const Value: string);
    procedure UpdateHeight;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMSetFont(var Message: TWMSetFont); message WM_SETFONT;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure SetDataSet(Value: TDataSet);
    procedure SetField(Value: string);
    procedure SetTableIndex(Value: string);
    function  FindQueryField: Boolean;
    function  CharKeys(var CharCode: Char): Boolean;
    procedure HomeEndKeys(CharCode: Word; Shift: TShiftState);
    procedure DeleteKeys(CharCode: Word);
    procedure CursorInc(CursorPos: Integer; Incr: Integer);
    procedure CursorDec(CursorPos: Integer);
    procedure ArrowKeys(CharCode: Word; Shift: TShiftState);
    function  DeleteSelection(var Value: string; Offset: Integer; Len: Integer): Boolean;
    function  IsInTable(CheckField: string): Boolean;
    procedure FindText;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CheckCursor;
    procedure SetCursor(Pos: Integer);
    procedure Change; dynamic;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure DoSetMaxLength(Value: Integer); virtual;
    function GetSelLength: Integer; virtual;
    function GetSelStart: Integer; virtual;
    function GetSelText: string; virtual;
    procedure SetSelLength(Value: Integer); virtual;
    procedure SetSelStart(Value: Integer); virtual;
    property AutoSelect: Boolean read FAutoSelect write FAutoSelect default True;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property CharCase: TEditCharCase read FCharCase write SetCharCase default ecNormal;
    property HideSelection: Boolean read FHideSelection write SetHideSelection default True;
    property MaxLength: Integer read FMaxLength write SetMaxLength default 0;
    property OEMConvert: Boolean read FOEMConvert write SetOEMConvert default False;
    property PasswordChar: Char read FPasswordChar write SetPasswordChar default #0;
    property ParentColor default False;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly default False;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnFind: TFindEvent read FFindEvent write FFindEvent;
  public
    procedure GetSel(var SelStart: Integer; var SelStop: Integer);
    procedure SetSel(SelStart: Integer; SelStop: Integer);
    constructor Create(AOwner: TComponent); override;
    procedure Clear; virtual;
    procedure ClearSelection;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure PasteFromClipboard;
    function GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer; virtual;
    procedure SelectAll;
    procedure SetSelTextBuf(Buffer: PChar);
    property Modified: Boolean read GetModified write SetModified;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelText: string read GetSelText write SetSelText;
    property Text;
    property TableIndex: string read FTableIndex write SetTableIndex;
    property Field: string read FField write SetField;
    property DataSet: TDataSet read FDataSet write SetDataSet;
    property NextRecType: TNextRecType read FNextRecType write FNextRecType default nrNextRecord;
  published
    property TabStop default True;
  end;

  TFindEdit = class(TCustomFindEdit)
  published
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property TableIndex;
    property Field;
    property DataSet;
    property NextRecType;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnFind;
  end;

  TFieldProperty = class (TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TIndexProperty = class (TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('A-Team', [TFindEdit]);
  RegisterPropertyEditor(TypeInfo(string),TFindEdit,'Field',TFieldProperty);
  RegisterPropertyEditor(TypeInfo(string),TFindEdit,'TableIndex',TIndexProperty);
end;

type
  TSelection = record
    StartPos, EndPos: Integer;
  end;

const
  BorderStyles: array[TBorderStyle] of Longint = (0, WS_BORDER);

function HasPopup(Control: TControl): Boolean;
begin
  Result := True;
  while Control <> nil do
    if TCustomFindEdit(Control).PopupMenu <> nil then Exit else Control := Control.Parent;
  Result := False;
end;

{ TCustomFindEdit }

constructor TCustomFindEdit.Create(AOwner: TComponent);
const
  EditStyle = [csClickEvents, csSetCaption, csDoubleClicks, csFixedHeight];
begin
  inherited Create(AOwner);
  if NewStyleControls then
    ControlStyle := EditStyle else
    ControlStyle := EditStyle + [csFramed];
  Width := 121;
  Height := 25;
  TabStop := True;
  ParentColor := False;
  FBorderStyle := bsSingle;
  FAutoSize := True;
  FAutoSelect := True;
  FHideSelection := True;
  AdjustHeight;
  FIsFind:=False;
end;

procedure TCustomFindEdit.SetDataSet(Value: TDataSet);
begin
  if FDataSet<>Value then FDataSet:=Value;
end;

procedure TCustomFindEdit.SetField(Value: string);
begin
  if FField<>Value then FField:=Value;
end;

procedure TCustomFindEdit.SetTableIndex(Value: string);
begin
  if FTableIndex<>Value then FTableIndex:=Value;
end;

procedure TCustomFindEdit.DoSetMaxLength(Value: Integer);
begin
  SendMessage(Handle, EM_LIMITTEXT, Value, 0)
end;

procedure TCustomFindEdit.SetAutoSize(Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateHeight;
  end;
end;

procedure TCustomFindEdit.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    UpdateHeight;
    RecreateWnd;
  end;
end;

procedure TCustomFindEdit.SetCharCase(Value: TEditCharCase);
begin
  if FCharCase <> Value then
  begin
    FCharCase := Value;
    RecreateWnd;
  end;
end;

procedure TCustomFindEdit.SetHideSelection(Value: Boolean);
begin
  if FHideSelection <> Value then
  begin
    FHideSelection := Value;
    RecreateWnd;
  end;
end;

procedure TCustomFindEdit.SetMaxLength(Value: Integer);
begin
  if FMaxLength <> Value then
  begin
    FMaxLength := Value;
    if HandleAllocated then DoSetMaxLength(Value);
  end;
end;

procedure TCustomFindEdit.SetOEMConvert(Value: Boolean);
begin
  if FOEMConvert <> Value then
  begin
    FOEMConvert := Value;
    RecreateWnd;
  end;
end;

function TCustomFindEdit.GetModified: Boolean;
begin
  Result := FModified;
  if HandleAllocated then Result := SendMessage(Handle, EM_GETMODIFY, 0, 0) <> 0;
end;

procedure TCustomFindEdit.SetModified(Value: Boolean);
begin
  if HandleAllocated then
    SendMessage(Handle, EM_SETMODIFY, Byte(Value), 0) else
    FModified := Value;
end;

procedure TCustomFindEdit.SetPasswordChar(Value: Char);
begin
  if FPasswordChar <> Value then
  begin
    FPasswordChar := Value;
    if HandleAllocated then
    begin
      SendMessage(Handle, EM_SETPASSWORDCHAR, Ord(FPasswordChar), 0);
      SetTextBuf(PChar(Text));
    end;
  end;
end;

procedure TCustomFindEdit.SetReadOnly(Value: Boolean);
begin
  if FReadOnly <> Value then
  begin
    FReadOnly := Value;
    if HandleAllocated then
      SendMessage(Handle, EM_SETREADONLY, Ord(Value), 0);
  end;
end;

function TCustomFindEdit.GetSelStart: Integer;
begin
  SendMessage(Handle, EM_GETSEL, Longint(@Result), 0);
end;

procedure TCustomFindEdit.SetSelStart(Value: Integer);
begin
  SendMessage(Handle, EM_SETSEL, Value, Value);
end;

function TCustomFindEdit.GetSelLength: Integer;
var
  Selection: TSelection;
begin
  SendMessage(Handle, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
  Result := Selection.EndPos - Selection.StartPos;
end;

procedure TCustomFindEdit.SetSelLength(Value: Integer);
var
  Selection: TSelection;
begin
  SendMessage(Handle, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
  Selection.EndPos := Selection.StartPos + Value;
  SendMessage(Handle, EM_SETSEL, Selection.StartPos, Selection.EndPos);
  SendMessage(Handle, EM_SCROLLCARET, 0,0);
end;

procedure TCustomFindEdit.Clear;
begin
  SetWindowText(Handle, '');
end;

procedure TCustomFindEdit.ClearSelection;
begin
  SendMessage(Handle, WM_CLEAR, 0, 0);
end;

procedure TCustomFindEdit.CopyToClipboard;
begin
  SendMessage(Handle, WM_COPY, 0, 0);
end;

procedure TCustomFindEdit.CutToClipboard;
begin
  SendMessage(Handle, WM_CUT, 0, 0);
end;

procedure TCustomFindEdit.PasteFromClipboard;
begin
  SendMessage(Handle, WM_PASTE, 0, 0);
end;

procedure TCustomFindEdit.SelectAll;
begin
  SendMessage(Handle, EM_SETSEL, 0, -1);
end;

function TCustomFindEdit.GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
var
  P: PChar;
  StartPos: Integer;
begin
  StartPos := GetSelStart;
  Result := GetSelLength;
  P := StrAlloc(GetTextLen + 1);
  try
    GetTextBuf(P, StrBufSize(P));
    if Result >= BufSize then Result := BufSize - 1;
    StrLCopy(Buffer, P + StartPos, Result);
  finally
    StrDispose(P);
  end;
end;

procedure TCustomFindEdit.SetSelTextBuf(Buffer: PChar);
begin
  SendMessage(Handle, EM_REPLACESEL, 0, LongInt(Buffer));
end;

function TCustomFindEdit.GetSelText: string;
var
  P: PChar;
  SelStart, Len: Integer;
begin
  SelStart := GetSelStart;
  Len := GetSelLength;
  SetString(Result, PChar(nil), Len);
  if Len <> 0 then
  begin
    P := StrAlloc(GetTextLen + 1);
    try
      GetTextBuf(P, StrBufSize(P));
      Move(P[SelStart], Pointer(Result)^, Len);
    finally
      StrDispose(P);
    end;
  end;
end;

procedure TCustomFindEdit.SetSelText(const Value: String);
begin
  SendMessage(Handle, EM_REPLACESEL, 0, Longint(PChar(Value)));
end;

procedure TCustomFindEdit.CreateParams(var Params: TCreateParams);
const
  Passwords: array[Boolean] of Longint = (0, ES_PASSWORD);
  ReadOnlys: array[Boolean] of Longint = (0, ES_READONLY);
  CharCases: array[TEditCharCase] of Longint = (0, ES_UPPERCASE, ES_LOWERCASE);
  HideSelections: array[Boolean] of Longint = (ES_NOHIDESEL, 0);
  OEMConverts: array[Boolean] of Longint = (0, ES_OEMCONVERT);
begin
  inherited CreateParams(Params);
  CreateSubClass(Params, 'EDIT');
  with Params do
  begin
    Style := Style or (ES_AUTOHSCROLL or ES_AUTOVSCROLL) or
      BorderStyles[FBorderStyle] or Passwords[FPasswordChar <> #0] or
      ReadOnlys[FReadOnly] or CharCases[FCharCase] or
      HideSelections[FHideSelection] or OEMConverts[FOEMConvert];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TCustomFindEdit.CreateWnd;
begin
  FCreating := True;
  try
    inherited CreateWnd;
  finally
    FCreating := False;
  end;
  SendMessage(Handle, EM_LIMITTEXT, FMaxLength, 0);
  Modified := FModified;
  if FPasswordChar <> #0 then
    SendMessage(Handle, EM_SETPASSWORDCHAR, Ord(FPasswordChar), 0);
  UpdateHeight;
end;

procedure TCustomFindEdit.DestroyWnd;
begin
  FModified := Modified;
  inherited DestroyWnd;
end;

procedure TCustomFindEdit.UpdateHeight;
begin
  if FAutoSize and (BorderStyle = bsSingle) then
  begin
    ControlStyle := ControlStyle + [csFixedHeight];
    AdjustHeight;
  end else
    ControlStyle := ControlStyle - [csFixedHeight];
end;

procedure TCustomFindEdit.AdjustHeight;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  if NewStyleControls then
  begin
    if Ctl3D then I := 8 else I := 6;
    I := GetSystemMetrics(SM_CYBORDER) * I;
  end else
  begin
    I := SysMetrics.tmHeight;
    if I > Metrics.tmHeight then I := Metrics.tmHeight;
    I := I div 4 + GetSystemMetrics(SM_CYBORDER) * 4;
  end;
  Height := Metrics.tmHeight + I;
end;

procedure TCustomFindEdit.Change;
begin
  inherited Changed;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TCustomFindEdit.DefaultHandler(var Message);
begin
  case TMessage(Message).Msg of
    WM_RBUTTONUP:
      if HasPopup(Self) then Exit;
    WM_SETFOCUS:
      if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) and
        not IsWindow(TWMSetFocus(Message).FocusedWnd) then
        TWMSetFocus(Message).FocusedWnd := 0;
  end;
  inherited;
end;

procedure TCustomFindEdit.WMSetFont(var Message: TWMSetFont);
begin
  inherited;
  if NewStyleControls and
    (GetWindowLong(Handle, GWL_STYLE) and ES_MULTILINE = 0) then
    SendMessage(Handle, EM_SETMARGINS, EC_LEFTMARGIN or EC_RIGHTMARGIN, 0);
end;

procedure TCustomFindEdit.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then
  begin
    UpdateHeight;
    RecreateWnd;
  end;
  inherited;
end;

procedure TCustomFindEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if (csFixedHeight in ControlStyle) and not ((csDesigning in
    ComponentState) and (csLoading in ComponentState)) then AdjustHeight;
end;

procedure TCustomFindEdit.CNCommand(var Message: TWMCommand);
begin
  if (Message.NotifyCode = EN_CHANGE) and not FCreating then Change;
end;

procedure TCustomFindEdit.CMEnter(var Message: TCMGotFocus);
begin
  FNewText:='';
  if FAutoSelect and not (csLButtonDown in ControlState) and
    (GetWindowLong(Handle, GWL_STYLE) and ES_MULTILINE = 0) then SelectAll;
  inherited;
end;

procedure TCustomFindEdit.CMTextChanged(var Message: TMessage);
begin
  inherited;
  if not HandleAllocated or (GetWindowLong(Handle, GWL_STYLE) and
    ES_MULTILINE <> 0) then Change;
end;

procedure TCustomFindEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key <> 0) and not (ssAlt in Shift) then
  begin
    if (Key = VK_LEFT) or(Key = VK_RIGHT) then
    begin
      ArrowKeys(Key, Shift);
      if not ((ssShift in Shift) or (ssCtrl in Shift)) then Key := 0;
      FIsFind:=False;
      Exit;
    end
    else if (Key = VK_UP) or(Key = VK_DOWN) then
    begin

    if (Key = VK_DOWN) and not FDataSet.EOF then
      begin
      if (FDataSet is TTable) then
        if (FDataSet as TTable).IndexName<>FTableIndex then (FDataSet as TTable).IndexName:=FTableIndex;
      if FNextRecType=nrNextValue then
        while (FDataSet.FieldByName(FField).AsString=Text) and (not FDataSet.EOF) do FDataSet.Next else FDataSet.Next;
      Text:=FDataSet.FieldByName(FField).AsString;
      FPosition:=0;
      SetCursor(FPosition);
      SetSel(FPosition,Length(Text));
      FIsFind:=True;
      if Assigned(FFindEvent) then FFindEvent(self,FIsFind);
      end;
    if (Key = VK_UP) and not FDataSet.BOF then
      begin
      if (FDataSet is TTable) then
        if (FDataSet as TTable).IndexName<>FTableIndex then (FDataSet as TTable).IndexName:=FTableIndex;
      if FNextRecType=nrNextValue then
        while (FDataSet.FieldByName(FField).AsString=Text) and (not FDataSet.BOF) do FDataSet.Prior else FDataSet.Prior;
      Text:=FDataSet.FieldByName(FField).AsString;
      FPosition:=0;
      SetCursor(FPosition);
      SetSel(FPosition,Length(Text));
      FIsFind:=True;
      if Assigned(FFindEvent) then FFindEvent(self,FIsFind);
      end;

    FNewText:='';  
    Key := 0;
//      FIsFind:=False;
    Exit;
    end
    else if (Key = VK_HOME) or(Key = VK_END) then
    begin
      HomeEndKeys(Key, Shift);
      Key := 0;
      FIsFind:=False;
      Exit;
    end
    else if ((Key = VK_DELETE) and ([ssShift, ssCtrl] * Shift = [])) or (Key = VK_BACK) then
    begin
      DeleteKeys(Key);
      Key := 0;
      FIsFind:=False;
      Exit;
    end;
    CheckCursor;
  end;
end;

procedure TCustomFindEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
  if (Key <> 0) then
  begin
    if ((Key = VK_LEFT) or(Key = VK_RIGHT)) and (ssCtrl in Shift) then
      CheckCursor;
  end;
end;

procedure TCustomFindEdit.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if (Key <> #0) and not (Char(Key) in [^V, ^X, ^C]) then
  begin
    CharKeys(Key);
    Key := #0;
  end;
end;

procedure TCustomFindEdit.CheckCursor;
var
  SelStart, SelStop: Integer;
begin
  if not HandleAllocated then  Exit;
  GetSel(SelStart, SelStop);
  if SelStart = SelStop then SetCursor(SelStart);
end;

function TCustomFindEdit.CharKeys(var CharCode: Char): Boolean;
var
  SelStart, SelStop : Integer;
  Txt: string;
  CharMsg: TMsg;
begin
  Result := False;
  if Word(CharCode) = VK_ESCAPE then
  begin
//    Reset;
    Exit;
  end;
  if ReadOnly then Exit;
  if (Word(CharCode) = VK_BACK) then Exit;
  if (Word(CharCode) = VK_RETURN) then
  begin
//    ValidateEdit;
    Exit;
  end;

  if FIsFind then
    begin
    Text:=FNewText;
    SetCursor(FPosition);
    end;

  GetSel(SelStart, SelStop);
  if (SelStop - SelStart) > 1 then
    begin
    DeleteKeys(VK_DELETE);
    SetCursor(SelStart);
    end;

  if (CharCode in LeadBytes) then
    if PeekMessage(CharMsg, Handle, WM_CHAR, WM_CHAR, PM_REMOVE) then
      if CharMsg.Message = WM_Quit then PostQuitMessage(CharMsg.wparam);
  if (CharCode in LeadBytes) then
    begin
    Txt := CharCode + Char(CharMsg.wParam);
    SetSel(SelStart, SelStart + 2);
    end
  else Txt := CharCode;
  SendMessage(Handle, EM_REPLACESEL, 0, LongInt(PChar(Txt)));
  GetSel(SelStart, SelStop);
  CursorInc(SelStart, 0);
  FindText;
end;

procedure TCustomFindEdit.FindText;
var SelStop: Integer;
begin
  if IsInTable(FField) then
    begin
    if (FDataSet is TTable) then
      begin
      if (FDataSet as TTable).IndexName<>FTableIndex then (FDataSet as TTable).IndexName:=FTableIndex;
      (FDataSet as TTable).SetKey;
      FDataSet.FieldByName(FField).AsString:=Text;
      (FDataSet as TTable).GotoNearest;
      FNewText:=Text;
      if FNewText=LeftStr(FDataSet.FieldByName(FField).AsString,Length(Trim(FNewText))) then
        begin
        GetSel(FPosition,SelStop);
        Text:=FDataSet.FieldByName(FField).AsString;
        SetCursor(FPosition);
        SetSel(FPosition,Length(Text));
        FIsFind:=True;
        end
      else
        begin
        Inc(FPosition);
        FIsFind:=False;
        end;
      if Assigned(FFindEvent) then FFindEvent(self,FIsFind);
      end
    else if (FDataSet is TQuery) then
      begin
      FDataSet.First;
      FNewText:=Text;
      if FindQueryField then
        begin
        GetSel(FPosition,SelStop);
        Text:=FDataSet.FieldByName(FField).AsString;
        SetCursor(FPosition);
        SetSel(FPosition,Length(Text));
        FIsFind:=True;
        end
      else
        begin
        Inc(FPosition);
        FIsFind:=False;
        end;
      end;
    end;
end;

function TCustomFindEdit.FindQueryField: Boolean;
begin
  Result:=False;
  while not FDataSet.EOF do
    if FNewText=LeftStr(FDataSet.FieldByName(FField).AsString,Length(FNewText)) then
      begin
      Result:=True;
      Break;
      end
    else FDataSet.Next;
end;

function TCustomFindEdit.IsInTable(CheckField: string): Boolean;
var i: Integer;
begin
  Result:=False;
  if Assigned(FDataSet) and FDataSet.Active then
    begin
    for i:=0 to FDataSet.FieldCount-1 do
      if UpperCase(CheckField)=FDataSet.Fields[i].FieldName then
        begin
        Result:=True;
        Break;
        end;
    end;
  if result and (FDataSet is TTable) then
    begin
    Result:=False;
    (DataSet as TTable).IndexDefs.Update;
    for i:=0 to (DataSet as TTable).IndexDefs.Count-1 do
     if UpperCase(FTableIndex)=(FDataSet as TTable).IndexDefs.Items[i].Name then
       begin
       Result:=True;
       Break;
       end;
    end;
end;

procedure TCustomFindEdit.GetSel(var SelStart: Integer; var SelStop: Integer);
begin
  SendMessage(Handle, EM_GETSEL, Integer(@SelStart), Integer(@SelStop));
end;

procedure TCustomFindEdit.SetSel(SelStart: Integer; SelStop: Integer);
begin
  SendMessage(Handle, EM_SETSEL, SelStart, SelStop);
end;

procedure TCustomFindEdit.SetCursor(Pos: Integer);
var
  SelStart, SelStop: Integer;
  KeyState: TKeyboardState;
  NewKeyState: TKeyboardState;
  I: Integer;
begin
  if ByteType(Text, Pos+1) = mbTrailByte then Dec(Pos);
  SelStart := Pos;
  if SelStart < 0 then SelStart := 0;
  SelStop  := SelStart {+ 1};
  if (Length(Text) > SelStop) and (Text[SelStop] in LeadBytes) then Inc(SelStop);
  if SelStart >= Length(Text) then
    begin
    SelStart := Length(Text);
    SelStop  := SelStart;
    end;
  SetSel(SelStop, SelStop);
  if SelStart <> SelStop then
    begin
    GetKeyboardState(KeyState);
    for I := Low(NewKeyState) to High(NewKeyState) do NewKeyState[I] := 0;
    NewKeyState [VK_SHIFT] := $81;
    NewKeyState [VK_LEFT] := $81;
    SetKeyboardState(NewKeyState);
    SendMessage(Handle, WM_KEYDOWN, VK_LEFT, 1);
    SendMessage(Handle, WM_KEYUP, VK_LEFT, 1);
    SetKeyboardState(KeyState);
    end;
  FCaretPos := SelStart;
end;

procedure TCustomFindEdit.HomeEndKeys(CharCode: Word; Shift: TShiftState);
var
  SelStart, SelStop : Integer;
begin
  GetSel(SelStart, SelStop);
  if (CharCode = VK_HOME) then
    begin
    if (ssShift in Shift) then
      begin
      if (SelStart <> FCaretPos) and (SelStop <> (SelStart + 1)) then SelStop := SelStart + 1;
      SetSel(0, SelStop);
      CheckCursor;
      end
    else SetCursor(0);
    FCaretPos := 0;
    end
  else
    begin
    if (ssShift in Shift) then
      begin
      if (SelStop <> FCaretPos) and (SelStop <> (SelStart + 1)) then SelStart := SelStop - 1;
      SetSel(SelStart, Length(Text));
      CheckCursor;
      end
    else SetCursor(Length(Text));
    FCaretPos := Length(Text);
    end;
end;

procedure TCustomFindEdit.DeleteKeys(CharCode: Word);
var
  SelStart, SelStop : Integer;
  NuSelStart: Integer;
  Str: string;
begin
  if ReadOnly then Exit;
  GetSel(SelStart, SelStop);
  if ((SelStop - SelStart) <= 1) and (CharCode = VK_BACK) then
    begin
    NuSelStart := SelStart;
    CursorDec(SelStart);
    GetSel(SelStart, SelStop);
    if SelStart = NuSelStart then Exit;
    end;
  Inc(SelStop);
//  if (SelStop - SelStart) < 1 then Exit;
//  Str := Text;
//  DeleteSelection(Str, SelStart, SelStop - SelStart);
  Str := '';//Copy(Str, SelStart+1, SelStop - SelStart);
  SetSel(SelStart, SelStop);
  if FIsFind then Text:=FNewText;
  SendMessage(Handle, EM_REPLACESEL, 0, LongInt(PChar(Str)));
  if (SelStop - SelStart) <> 1 then
    begin
//    SelStart := ;
    SetCursor(SelStart);
    end
  else
    begin
    GetSel(SelStart, SelStop);
    SetCursor(SelStart);
    end;
//  FindText;
end;

procedure TCustomFindEdit.CursorInc(CursorPos: Integer; Incr: Integer);
var NuPos: Integer;
begin
  NuPos := CursorPos + Incr;
//  NuPos := GetNextEditChar(NuPos);
//  if IsLiteralChar(EditMask, nuPos) then
//    NuPos := CursorPos;
  SetCursor(NuPos);
end;


procedure TCustomFindEdit.CursorDec(CursorPos: Integer);
var nuPos: Integer;
begin
  nuPos := CursorPos;
  Dec(nuPos);
//  nuPos := GetPriorEditChar(nuPos);
  SetCursor(NuPos);
end;

function TCustomFindEdit.DeleteSelection(var Value: string; Offset: Integer; Len: Integer): Boolean;
var
  i, EndDel: Integer;
  StrOffset, MaskOffset, Temp: Integer;
//  CType: TMaskCharType;
begin
  Result := True;
  if Len = 0 then Exit;
  StrOffset := Offset + 1;
  EndDel := StrOffset + Len;
  for MaskOffset := Offset to Length(Text) do
    begin
    Value[StrOffset] := ' ';
    Inc(StrOffset);
    if StrOffset >= EndDel then Break;
    end;
end;

procedure TCustomFindEdit.ArrowKeys(CharCode: Word; Shift: TShiftState);
var SelStart, SelStop : Integer;
begin
  if (ssCtrl in Shift) then Exit;
  GetSel(SelStart, SelStop);
  if (ssShift in Shift) then
    begin
    if (CharCode = VK_RIGHT) then
      begin
      Inc(FCaretPos);
      if (SelStop = SelStart + 1) then
        begin
        SetSel(SelStart, SelStop);  {reset caret to end of string}
        Inc(FCaretPos);
        end;
      if FCaretPos > Length(Text) then FCaretPos := Length(Text);
      end
    else  {if (CharCode = VK_LEFT) then}
      begin
      Dec(FCaretPos);
      if (SelStop = SelStart + 2) and (FCaretPos > SelStart) then
        begin
        SetSel(SelStart + 1, SelStart + 1);  {reset caret to show up at start}
        Dec(FCaretPos);
        end;
      if FCaretPos < 0 then FCaretPos := 0;
      end;
    end
  else
    begin
    if (SelStop - SelStart) > 1 then
      begin
      if ((SelStop - SelStart) = 2) and (Text[SelStart+1] in LeadBytes) then
        begin
        if (CharCode = VK_LEFT) then CursorDec(SelStart) else CursorInc(SelStart, 2);
        Exit;
        end;
      if SelStop = FCaretPos then Dec(FCaretPos);
      SetCursor(FCaretPos);
      end
    else if (CharCode = VK_LEFT) then CursorDec(SelStart)
    else   { if (CharCode = VK_RIGHT) then  }
      begin
      if SelStop = SelStart then SetCursor(SelStart+1)
      else if Text[SelStart+1] in LeadBytes then CursorInc(SelStart, 2) else CursorInc(SelStart, 1);
      end;
    end;
end;

procedure TCustomFindEdit.WMLButtonUp(var Message: TWMLButtonUp);
var SelStart, SelStop : Integer;
begin
  inherited;
  FIsFind:=False;
  GetSel(SelStart, SelStop);
  FCaretPos := SelStart;
  if (SelStart <> SelStop) and (Message.XPos > FBtnDownX) then FCaretPos := SelStop;
  CheckCursor;
end;

procedure TCustomFindEdit.WMLButtonDown(var Message: TWMLButtonDown);
begin
  inherited;
  FIsFind:=False;
  FBtnDownX := Message.XPos;
end;


function TFieldProperty.GetAttributes: TPropertyAttributes;
begin
  Result:=[paValueList];
end;

procedure TFieldProperty.GetValues(Proc: TGetStrProc);
var i: Integer;
    tmpFindEdit: TFindEdit;
begin
    tmpFindEdit:=GetComponent(0) as TFindEdit;
    if Assigned(tmpFindEdit.DataSet) then if tmpFindEdit.DataSet.Active then
      for i:=0 to tmpFindEdit.DataSet.FieldCount-1 do Proc(tmpFindEdit.DataSet.Fields[i].FieldName);
end;

function TIndexProperty.GetAttributes: TPropertyAttributes;
begin
  Result:=[paValueList];
end;

procedure TIndexProperty.GetValues(Proc: TGetStrProc);
var i: Integer;
    tmpFindEdit: TFindEdit;
begin
    tmpFindEdit:=GetComponent(0) as TFindEdit;
    if tmpFindEdit.DataSet is TTable then
      begin
      with tmpFindEdit do
        if Assigned(DataSet) then if DataSet.Active then
          begin
          (DataSet as TTable).IndexDefs.Update;
          for i:=0 to (DataSet as TTable).IndexDefs.Count-1 do Proc((DataSet as TTable).IndexDefs.Items[i].Name);
          end;
      end;
end;


end.
